module ASTPlus (
  Statement (..)
  , StatementF (..)
  , isIsomorphic 
  , UnaOp (..)
  , BinOp (..)
  , Arg (..)
  , StatementPlus (..)
  , StatementPlusF (..)
  , VarName
  , extractPlus 
  , extractPlusF
  , promoteTree
  , annotateTreeV 
  , extractDefs 
  , groupDef 
  , treeGenKillV 
  , treeWalker
  , mapTree
  , groupEquiv
  , getExpressionVars
  , mapTreeValV
  , makeExampleAST
  , makeExampleAST'
  , getLeavesV
  , enumerateTreeV
  , flattenVarList
  ) where
       
-- import Data.String
import Data.List
import Data.Fix
import qualified Data.Set as Set

import Expression
import DMSUtil


type VarName = String
data StatementF x = ExpressionStm Expression
                 | CompoundStm Statement x
                 | AssignmentStm VarName Expression
                 | IfStm Expression x x
                 | DoStm Expression x
                 deriving (Eq, Show, Ord)
                          
type Statement = Fix (StatementF)
                        
-- A polymorphic type for statements augmented with an additional term
data StatementPlusF a x = ExpStPlus a Expression 
                        | CompoundPlus a x x
                        | AssignPlus a VarName Expression
                        | IfPlus a Expression x x
                        | DoPlus a Expression x
                        deriving (Eq, Ord, Show)

type StatementPlus a = Fix (StatementPlusF a)


{- Everything else works on augmented trees, so we provide a function to promote
regular trees to augmented form.  val can and probably should be (). -}

promoteTree s val =
  case s of
    Fix (ExpressionStm e) -> 
      Fix $ ExpStPlus val e 
    Fix (CompoundStm l r) -> 
      Fix $ CompoundPlus val (promoteTree l val) (promoteTree r val)
    Fix (AssignmentStm n v) -> 
      Fix $ AssignPlus val n v
    Fix (IfStm e th el) ->  
      Fix $ IfPlus val e (promoteTree th val) (promoteTree el val)
    Fix (DoStm t b) -> 
      Fix $ DoPlus val t (promoteTree b val)

extractPlusF s = 
  case s of
    ExpStPlus a _ -> a
    CompoundPlus a _ _ -> a
    AssignPlus a _ _   -> a
    IfPlus a _ _ _     -> a
    DoPlus a _ _       -> a


extractPlus :: StatementPlus a -> a
extractPlus s =
  case s of
    Fix (ExpStPlus a _) -> a
    Fix (CompoundPlus a _ _) -> a
    Fix (AssignPlus a _ _)   -> a
    Fix (IfPlus a _ _ _)     -> a
    Fix (DoPlus a _ _)       -> a

-- if you have an AST annotated with numbering of assignment statements,
-- this will extract a proplist of the numbers and targets of assignments.
-- Note: not currently used!
-- (Superceded by the graph-based approach defined in CFG module.)    
extractDefs :: StatementPlus Int -> [(Int, String)]
extractDefs t =
  case t of 
    Fix (ExpStPlus _ e)      -> []
    Fix (CompoundPlus _ l r) -> (extractDefs l) ++ (extractDefs r)
    Fix (AssignPlus i n v)   -> [(i, n)]
    Fix (IfPlus _ _ th e)    -> (extractDefs th) ++ (extractDefs e)
    Fix (DoPlus _ _ b)       -> extractDefs b   

-- Partition definitions into sets that define the same variable
-- Note: not currently used!
groupDef defs  = 
  let
    -- the compiler complains if I don't nail this type down.
    sameVar :: (b, String) -> (b, String) -> Bool
    sameVar (_, s1) (_, s2) = (s1==s2)
    processDefGroup xs =
      let var = snd $ head xs
      in (var, map fst xs)
    groups = groupEquiv sameVar defs
  in
   map processDefGroup groups

-- NB: find ((==2).fst) [(1, 2), (2,3)] -> Just 2, 3
-- Note: not currently used!
getValues k defGroups = 
  case find ((==k) . fst) defGroups of
    Just (k, v) -> v
    Nothing -> []



mapTree :: (a->b) -> (StatementPlus a) -> (StatementPlus b)
mapTree f t =
  let (_, r) = treeWalker (\_ p -> ((), f $ extractPlusF p)) () t
  in r

-- Abstract out tree traversal; the visitor function in the first argument does
-- the non-generic work.
-- We traverse the tree in post-order (i.e, depth first)
treeWalker :: (c -> StatementPlusF a b -> (c, b)) -> c -> StatementPlus a -> 
              (c, StatementPlus b)
-- The type signature is not spectacularly helpful.
-- c is the type of the accumulator; b is the type of the annotation of the output tree;
-- a is the type of the annotation of the input tree.
treeWalker transFn acc at =
  case at of
    Fix (ExpStPlus plus e) ->
      let (nacc, nplus) = transFn acc $ ExpStPlus plus e
      in (nacc, Fix $ ExpStPlus nplus e)
    Fix (CompoundPlus plus tl tr) ->
       let
         (acc1, ntl) = treeWalker transFn acc tl
         plus1 = extractPlus ntl
         (acc2, ntr) = treeWalker transFn acc1 tr
         plus2 = extractPlus ntr
         (nacc, nplus) = transFn acc2 $ CompoundPlus plus plus1 plus2
       in (nacc, Fix $ CompoundPlus nplus ntl ntr)
    Fix (AssignPlus plus name e) ->
      let
        (nacc, nplus) = transFn acc $ AssignPlus plus name e
      in (nacc, Fix $ AssignPlus nplus name e)
    Fix (IfPlus plus e tl tr) ->
      let
        (acc1, ntl) = treeWalker transFn acc tl
        plus1 = extractPlus ntl
        (acc2, ntr) = treeWalker transFn acc1 tr
        plus2 = extractPlus ntr
        (nacc, nplus) = transFn acc2 $ IfPlus plus e plus1 plus2
      in (nacc, Fix $ IfPlus nplus e ntl ntr)
    Fix (DoPlus plus e t)     ->
      let
        (acc1, nt) = treeWalker transFn acc t
        plus1 = extractPlus nt
        (nacc, nplus) = transFn acc1 $ DoPlus plus e plus1
      in (nacc, Fix $ DoPlus nplus e nt)


-- The machinery was a bunch of work, but this is cute:
-- we define how to combine gen and kill sets for various nodes,
-- then we can just hand this to our tree-walker.
--
-- NB: this is not actually used in the current cfg code, which gives
-- me a sad in the feelings but what can you do?
treeGenKillV defGrp t = 
  let 
    g name = getValues name defGrp
    transFn acc t =
      case t of
        ExpStPlus _ e ->
          (acc, (Set.empty, Set.empty))
        CompoundPlus _ l r ->
          let
            (gen1, kill1) = l
            (gen2, kill2) = r
            gen = gen2 `Set.union` (gen1 Set.\\ kill2)
            kill = kill2 `Set.union` (kill1 Set.\\ gen2)
          in (acc, (gen, kill))
        AssignPlus defno name e ->
          let
            gen = Set.singleton defno
            kill = Set.fromList (g name) Set.\\ gen
          in (acc, (gen, kill))
        IfPlus _ e l r ->
          let 
            (gen1, kill1) = l
            (gen2, kill2) = r
            gen = gen1 `Set.union` gen2
            kill = kill1 `Set.intersection` kill2
          in (acc, (gen, kill))
        DoPlus _ e v -> 
          let (gen, kill) =  v
          in (acc, (gen, kill))
  in
   treeWalker transFn () t

-- This function is used to annotate trees with definition numbers;
-- All other statements are labelled with 0.  (If I was hardcore, we
-- could have Just Int for the assignments and Nothing for the
-- others.)
annotateTreeV s =
  let
    transFn :: Integer -> StatementPlusF a Integer -> (Integer, Integer)
    transFn i t =
        -- as required, first element is accumulator, second is tree decoration.
        case t of 
          ExpStPlus _ e        -> (i, 0)
          CompoundPlus _ l r   -> (i, 0) 
          AssignPlus _ n v     -> (i+1, i)
          IfPlus _ te th e     -> (i, 0)
          DoPlus _ e b         -> (i, 0)
    (_, res) = treeWalker transFn 0 s
  in res

enumerateTreeV :: StatementPlus a -> StatementPlus Integer
enumerateTreeV s = 
  let
    transFn :: Integer -> StatementPlusF a Integer -> (Integer, Integer)
    transFn i t = (i+1, i)
        -- as required, first element is accumulator, second is tree decoration.
    (_, res) = treeWalker transFn 0 s
  in res


mapTreeValV :: (a -> b) -> (StatementPlus a) -> (StatementPlus b)
mapTreeValV f s =
  let transFn _ t =
        -- as required, first element is accumulator, second is tree decoration.
        ((), f $ extractPlusF t)
      (_, res) = treeWalker transFn () s
  in res

getPostOrderV stm =
  let
    transFn :: [StatementPlusF a ()] -> StatementPlusF a () -> ([StatementPlusF a ()], ())
    transFn acc t =
      case t of
        CompoundPlus _ _ _ -> (acc, ())
        ; _              -> (t:acc, ())
    (acc, _) = treeWalker transFn [] stm
  in reverse acc

getLeavesV stm =
  -- could probably be written as a filter on getPostOrderV 
  -- which we'll need anyway.
  let
    transFn :: [StatementPlusF a ()] -> StatementPlusF a () -> ([StatementPlusF a ()], ())
    transFn acc t =
        case t of 
          ExpStPlus _ e      -> (t:acc, ())
          ; AssignPlus _ n v -> (t:acc, ())
          ; _              -> (acc, ())
    (acc, _) = treeWalker transFn [] stm
  in reverse acc

exampleIf = Fix (CompoundStm 
                 (Fix $ AssignmentStm "fn" $ StringExp "datafile")
                 (Fix (IfStm (VarExp "a")
                       (Fix $ CompoundStm
                        (Fix $ AssignmentStm "c" $ IntExp 1)
                        (Fix $ AssignmentStm "b" $ IntExp 0))
                       (Fix $ AssignmentStm "b" $ IntExp 2))))

makeExampleAST =
  Fix (CompoundStm 
       (Fix $ AssignmentStm "fn" $ StringExp "datafile")
       (Fix $ CompoundStm 
        (Fix $ IfStm (VarExp "a")
         (Fix $ CompoundStm
          (Fix $ AssignmentStm "c" $ IntExp 1)
          (Fix $ AssignmentStm "b" $ IntExp 0))
         (Fix $ AssignmentStm "b" $ IntExp 2))
        $ Fix $ CompoundStm (Fix $ DoStm (VarExp "b") 
                             (Fix $ AssignmentStm "b" $ VarExp "b")) 
        $ Fix $ CompoundStm  (Fix $ AssignmentStm "b2"
                              $ ProcCallExp "blob1"  [PosArg (FloatExp 0.7)
                                               , PosArg (StringExp "HALF")
                                               , PosArg (VarExp "b")])
        $ Fix $ CompoundStm 
        (Fix $ AssignmentStm "b3"
         $ ProcCallExp "blob2" [KeyArg "data" (VarExp "b2")
                               , KeyArg "opcode" (StringExp "CAL")])
        $ Fix $ AssignmentStm "b" $ FloatExp 2.0))

makeExampleAST' =
  Fix $ CompoundStm 
  (Fix $ AssignmentStm "fn" $ StringExp "datafile")
  $ Fix $ CompoundStm (Fix $ IfStm (VarExp "a")
                       (Fix $ CompoundStm
                        (Fix $ AssignmentStm "b" $ IntExp 0)
                        (Fix $ AssignmentStm "c" $ IntExp 1))
                       (Fix $ AssignmentStm "b" $ IntExp 2))
  $ Fix $ CompoundStm  
  (Fix $ AssignmentStm "b2"
   $ ProcCallExp "blob1"  [ PosArg (FloatExp 0.7)
                          , PosArg (StringExp "FAIL")
                          , PosArg (VarExp "b")])
  $ Fix $ CompoundStm 
  (Fix $ AssignmentStm "b3"
   $ ProcCallExp "blob2" [KeyArg "data" (VarExp "b2")
                         , KeyArg "opcode" (StringExp "CAL")])
  $ Fix $ AssignmentStm "b" $ FloatExp 3.0
