module Alpern (
  extractLabel
  , makeValueGraph
  , labelArguments
  , groupByFunctions
  , Label (..)
  , Edge (..)
  , Node (..) 
  , hopcroftPartition
  , isVarNode
  , isVarFromFundef
  )


where

import Data.Maybe
import Data.List as DL
import Data.Set as DS
import MyGraph (getNodes)
import FlowGraph
import Expression 
import DMSUtil
import Debug.Trace -- because I'm stuck
 
{- 
Recall that:

data BasicItem a =  ExpNode a
                 | DefNode String a
                 | PhiNode String a
                 deriving (Eq, Ord)

data Expression = ProcCallExp String [Arg]
                | VarExp String
                | FloatExp Double
                | IntExp Integer
                | UnaExp UnaOp Expression
                | BinExp BinOp Expression Expression
                | StringExp String
                deriving (Eq, Ord)
-}

data Label = IntL Integer
             | FloatL Double
             | FuncL String
             | StringL String
             | LookupL String
             | SubstL String String
             | OtherL
             deriving (Show, Eq)

-- used in value graphs:
type ArgLabel = String
type NodeNum = Integer
type Definee = Maybe String

type Edge = (ArgLabel, NodeNum, NodeNum)
type Node = (NodeNum, Definee, Label)

isVarNode (_, m, _) = isJust m

isVarFromFundef(_, Just _, FuncL _) = True
isVarFromFundef(_, _, _)            = False

extractLabel e =
   case e of
     FloatExp d -> FloatL d
     IntExp i -> IntL i
     StringExp s -> StringL s
     ProcCallExp f _ -> FuncL f
     BinExp o _ _ -> FuncL (show o)
     UnaExp o _ -> FuncL (show o)
     VarExp n -> LookupL n

makeValueGraph cfg i0 = 
  let
    items = DL.concat $ DL.map (extractItems . snd) $ getNodes cfg
    (i, vges, vgns) = DL.foldl doNode (i0, [], []) items
    (vgns1, vges1) = substInVars vgns vges
    vges2 = groupEdges vges1
  in 
   (i, vges2, vgns1)

doNode (i, edges, nodes) n =
  let 
    definee = extractItemDefVar n :: Maybe String
    exp =  extractItemExpression n 
    (j, edges', nodes') = recurseExpression (i, edges, nodes) exp definee
  in
   (j, edges', nodes')
        
recurseExpression :: (Integer, [Edge], [Node]) -> Expression -> Maybe String ->
                     (Integer, [Edge], [Node])
recurseExpression (i, edges, nodes) e definee =
  let l = extractLabel e
      nodes0 = (i, definee, l):nodes
      j0 = i+1
      -- the rules: you insert yourself as node i; you pass on i+1 to
      -- your first child.
  in case e of
    (ProcCallExp n args) ->
      let lArgs = labelArguments args
          accn = DL.foldr (\(l,e) (j', edges', nodes') ->
                            let (j'', edges'', nodes'') =
                                  recurseExpression (j', edges', nodes') e Nothing
                            in (j'', (l, i, j'):edges'', nodes''))
                 (j0, edges, nodes0) lArgs
      in accn
    (UnaExp op e) ->
      let (j, edges1, nodes1) =
            recurseExpression (j0, edges, nodes0) e Nothing
      in (i+1, ("0", i, i+1):edges1, nodes1)
    (BinExp op el er) ->
      let (j1, edges1, nodes1) =
            recurseExpression (j0, edges, nodes0) el Nothing
          (j2, edges2, nodes2) =
            recurseExpression (j1, edges1, nodes1) er Nothing
          -- j2 is the number of the node to be assigned next;
          -- j0 and j1 are our child nodes.
       in (j2, ("0", i, j0):("1", i, j1):edges2, nodes2)
    (StringExp _) -> (j0, edges, nodes0)
    (VarExp _)    -> (j0, edges, nodes0)
    (FloatExp _)  -> (j0, edges, nodes0)
    (IntExp _)    -> (j0, edges, nodes0)

labelArguments args =
  labelArgumentsAux 0 args []
  where
    labelArgumentsAux :: Integer -> [Arg] -> [(String, Expression)] ->
                         [(String, Expression)]
    labelArgumentsAux i ((PosArg e):args) res =
      labelArgumentsAux (i+1) args ((show i, e):res)
    labelArgumentsAux _ ((KeyArg k e):args) res =
      labelArgumentsAux (-1) args ((k, e):res)
    labelArgumentsAux _ [] res = reverse res

substInVars :: [Node] -> [Edge] -> ([Node], [Edge])
substInVars nodes0 edges0 =
  let defs = DL.filter (\(_, e, _) -> isJust e) nodes0
      plDefs = DL.map (\(n, Just v, _) -> (v, n)) defs
      edges' = DL.foldl (\es n -> replaceLookup n plDefs es) edges0 nodes0
      nodes1 = DL.filter notLookup nodes0
  in
   trace ("nodes0 " ++ show nodes0) $
   trace ("nodes1 " ++ show nodes1) $ 
   (nodes1, edges')
  where
    replaceLookup :: Node -> [(String, Integer)] -> [Edge] -> [Edge]
    replaceLookup (i, v0, LookupL v1) plDefs es =
      let j = lookupOrDie v1 plDefs 
      in DL.map (replaceEdge i j) es
    replaceLookup _  _ es = es
    replaceEdge :: Integer -> Integer -> Edge -> Edge
    replaceEdge i j e@(s, in_, out) =
      case i==out of
        True -> (s, in_, j)
        _   -> e
    notLookup (_, _, LookupL _) = False
    notLookup _ = True

groupEdges es =
  -- a smarter person would factor this out as a general algorithm.
  aux es []
  where
    aux (e@(index, _, _):es) sorted =
      let 
        (matches, remainder) = DL.partition (\(i1,_,_) -> i1 == index) es
        newGroup = DL.map (\(_, from, to) -> (from, to)) (e:matches)
      in aux remainder ((index, newGroup):sorted)
    aux [] result = result

-- the Haskell groupBy doesn't do this; it just groups *successive* elements,
-- which is not at all what I want.
myGroupBy fun es = 
  aux es []
  where
    aux (e1:es) sorted =
      -- NB: fun is a function of two arguments used in
      -- partially-applied curried form
      let (matches, remainder) = DL.partition (fun e1) es
      in aux remainder ((e1:matches):sorted)
    aux [] result = result

groupByFunctions ns =
  let groups = myGroupBy sameFun ns
      nodeGroups = DL.map (\g -> DL.map extractNode g) groups
  in
   nodeGroups
  where
    sameFun (_, _, l1) (_, _, l2) = (l1==l2)
    extractNode (n, _, _) = n

-- the partition algorithm is from Alpern figure 7; it is attributed to
-- Hopcroft. bs0 is the initial partition (into nodes with the same
-- function); fis is the map function-argument placeholders (either
-- stringified integers or keyword arguments) to inverted edges
hopcroftPartition bs0 fis =
  let
    q = (length bs0) - 1
    waiting0 = [0..q]
  in
   loop1 q bs0 waiting0 
  where
    loop1 _ bs [] = bs
    loop1 q bs (i:waiting) = 
      let
        b = bs !! i
        ms = DL.map fst fis
        (q', bs', waiting') = DL.foldl (loop2 b) (q, bs, waiting) ms
      in loop1 q' bs' waiting' 
    loop2 b (q, bs, waiting) m =
      let 
        fi = lookupOrDie m fis
        inverse = preImages b fi
      in trace ("Alpern fi " ++ show fi) $
         trace ("b " ++ show b) $
         trace ("Inverse " ++ show inverse) $
       DL.foldl (loop3 inverse) (q, bs, waiting) [0..q]
    loop3 inverse (q, bs, waiting) j =
      let bj = DS.fromList $ bs !! j
          inv = DS.fromList inverse
      in
       if (not $ DS.null (bj `DS.intersection` inv)) &&
          not (bj `DS.isSubsetOf` inv)
       then let
         q' = q+1
         bq =  bj `DS.intersection` inv
         bj' = bj DS.\\ bq
         bs' = (replaceIndex j (DS.toList bj') bs) ++ [DS.toList bq]
         waiting' =
           if j `elem` waiting then
             waiting ++ [q]
           else if (size bj) <= (size bq)
                then waiting ++ [j]
                else waiting ++ [q]
         in
          trace ("Alpern: splitting " ++ show bj ++ " inv " ++ show inv) $
          trace ("New bs " ++ show bj' ++ " & " ++ show bq) $
          (q', bs', waiting')
       else (q, bs, waiting)
    preImages b f =
      concat [lookupAll n f | n <- b]
      
  
