module Expression (
  Expression (..)
  , UnaOp (..)
  , BinOp (..)
  , Arg (..) 
  , flattenVarList
  , isIsomorphic
  , getExpressionVars
  ) where

import Data.List

data Expression = ProcCallExp String [Arg]
                | VarExp String
                | FloatExp Double
                | IntExp Integer
                | UnaExp UnaOp Expression
                | BinExp BinOp Expression Expression
                | StringExp String
                deriving (Eq, Ord)

data UnaOp = UnaMinus deriving (Show, Eq, Ord)

data BinOp = MLT | MGT | Plus | Minus | EqOp
           deriving (Eq, Ord)

instance Show BinOp where
  show MLT = "<"
  show MGT = ">"
  show Plus = "+"
  show Minus = "-"
  show EqOp = "=="

data Arg = PosArg Expression
          | KeyArg String Expression
          deriving (Eq, Ord)


flattenArgVarList (PosArg e) = flattenVarList e
flattenArgVarList (KeyArg _ e) = flattenVarList e


instance Show Arg where
  show (PosArg e) =
    show e
  show (KeyArg k e) =
    k ++ "=" ++ (show e)

getExpression (PosArg e) = e
getExpression (KeyArg _ e) = e


instance Show Expression where
  show e =
    case e of
      ProcCallExp name params ->
        let pt = [show p | p <- params]
        in name ++ "(" ++ (intercalate "," pt) ++ ")"
      ; VarExp s -> s
      ; FloatExp f -> show f
      ; IntExp i -> show i
      ; UnaExp o e -> (show o) ++ (show e)
      ; BinExp o l r -> (show l) ++ (show o) ++ (show r)
      ; StringExp s -> ['"'] ++ s ++ ['"']



flattenVarList :: Expression -> [String]
flattenVarList (VarExp n) = [n]
flattenVarList (FloatExp _) = []
flattenVarList (IntExp _) = []
flattenVarList (StringExp _) = []
flattenVarList (UnaExp _ e) = flattenVarList e
flattenVarList (BinExp _ l r) = (flattenVarList l) ++ (flattenVarList r)
flattenVarList (ProcCallExp _ ps) = concat $ map flattenArgVarList ps


getExpressionVars e =
  case e of
    VarExp n -> [n]
    ProcCallExp n params ->
      concat $ map (\p -> (getExpressionVars $ getExpression p)) params
    UnaExp _ e' -> getExpressionVars e'
    BinExp _ l r -> (getExpressionVars l) ++ (getExpressionVars r)
    otherwise -> []

isArgIsomorphic (PosArg e1) (PosArg e2) =
  isIsomorphic e1 e2
isArgIsomorphic (KeyArg n1 e1) (KeyArg n2 e2) =
  (n1 == n2) && (isIsomorphic e1 e2)

isIsomorphic (VarExp _) (VarExp _) = True
isIsomorphic (FloatExp v1) (FloatExp v2) = (v1 == v2)
isIsomorphic (IntExp v1) (IntExp v2) = (v1 == v2)
isIsomorphic (StringExp v1) (StringExp v2) = (v1 == v2)
isIsomorphic (UnaExp op1 e1) (UnaExp op2 e2) =
  (op1 == op2) && (isIsomorphic e1 e2)
isIsomorphic (BinExp op1 l1 r1) (BinExp op2 l2 r2) =
  -- Some operations may be commutative? We wouldn't know anything
  -- about that, sir or madam!
  (op1 == op2) && (isIsomorphic l1 r1) && (isIsomorphic l2 r2) 
isIsomorphic (ProcCallExp n1 as1) (ProcCallExp n2 as2) =
  (n1 == n2) && ((length as1) == (length as2))
  -- FIXME: the below is too paranoid: we don't allow reordered
  -- keyword arguments.
  && (and $ zipWith isArgIsomorphic as1 as2)
-- Everything else is false.
isIsomorphic _ _ = False

