module Tree (
  Tree (..)
  , StmNode (..)
  , parent
  , children
  , treeifyAST
  , postTreeMap
  , stripNodes
  , foldrTree
  , getChildren
  , countChildren
  , countNodes
  , getLabel
  , getLeaves
  , breadthFirst
  , preOrder
  , postOrder
  , showValue
  ) where


import Data.Fix
import ASTPlus 
import StmNode

{- We replace the old StatementPlus (AST) type with a new Tree type as
recommended by Fluri et al.  I'm starting to think this was a mistake,
but still.  Tree has nodes, each of which has a value and a *list* of
children, as opposed to named branches with child nodes. -}


-- extract a map of integers to nodes from an augmented AST.
data Tree a = Node a [Tree a]
            deriving Show

parent (Node p cs) = p
children (Node p cs) = cs


-- Convert fancy ASTs to less-fancy trees, with child-nodes arranged
-- in regular lists where possible.
treeifyAST :: StatementPlus Integer -> Tree (Integer, StmNode)
treeifyAST at =
  case at of
    Fix (ExpStPlus i e) -> Node (i, ExpStNode e) []
    Fix (CompoundPlus i tl tr) ->
      let
        children = mergeCompounds (treeifyAST tl) (treeifyAST tr)
      in
      Node (i, CompoundNode) children
    Fix (AssignPlus i name e) ->
      Node (i, AssignNode name e) []
    Fix (IfPlus i e tl tr) ->
      Node (i, IfNode e) [(treeifyAST tl), (treeifyAST tr)]
    Fix (DoPlus i e t)     ->
      Node (i, DoNode e) [treeifyAST t]
  where
    mergeCompounds l r =
      let
        l' = case l of 
          Node (_, CompoundNode) lst -> lst
          ; _ -> [l]
        r' = case r of
          Node (_, CompoundNode) lst -> lst
          ; _ -> [r]
      in l' ++ r'


postTreeMap :: (a -> b) -> Tree a -> Tree b
postTreeMap f (Node n children) =
  Node (f n) $ map (postTreeMap f) children

stripNodes :: Tree (Integer, StmNode) -> Tree Integer
stripNodes t = postTreeMap (\(i, _) -> i) t

foldrTree :: (a -> b -> b) -> b -> Tree a -> b
foldrTree f z (Node k (c:cs)) =
  foldrTree f (foldrTree f z (Node k cs)) c
foldrTree f z (Node k []) =
  f k z

countNodes (Node n children) z = foldrTree (\a z -> z+1) z

countPost t = foldrTree (\a (as, z) -> (a:as, z+1)) ([], 0) t 

getLeaves :: Tree a -> [a]
getLeaves (Node x ts) =
  case ts of
    [] -> [x]
    _ -> concat $ map getLeaves ts

breadthFirst t =
  breadthFirstAux [t] []
  where
    breadthFirstAux :: [Tree a] -> [a] -> [a]
    breadthFirstAux (t:ts) r =
      let
        (Node p cs) = t
      in 
       breadthFirstAux (ts++cs) (p:r)
    breadthFirstAux [] r = reverse r

-- This is more confusing than I would have expected.
countChildren :: Tree a -> [(a, Integer)]
countChildren (Node x children) =
  let cmaps = map countChildren children -- list contains all descendents for each child
      cTotal = sum $ map (snd . head) cmaps -- extract only direct children for sum
      counts =  concat cmaps -- now combine descendent results
      nChildren = (fromIntegral $ length children) + (cTotal)
  in (x, nChildren):counts

preOrder :: Tree a -> [a]
preOrder (Node x ts) =
  (:) x $ concat $ map postOrder ts

postOrder t = (reverse . preOrder) t

-- if you've got this, and preOrder and stuff, why do we work so hard
-- at countChildren?
getChildren t@(Node x1 ts) x =
  if x==x1 then
    (reverse . tail . preOrder) t 
  else
    concat $ map (\t -> getChildren t x) ts
