{- From 
https://hackage.haskell.org/package/xdot-0.1/docs/src/Graphics-XDot-Parser.html 
-}

import System.Environment
import System.IO
import qualified Data.GraphViz as D
import qualified Data.GraphViz.Attributes.Complete as A
import Data.GraphViz
import Data.GraphViz.Commands.IO
import qualified Data.Text.Lazy.IO as L
import qualified Data.Text.Lazy as DTL
import qualified Data.GraphViz.Types.Generalised as G
import Data.Maybe
import Data.Either
import Control.Monad

import Data.Graph as G
import Data.List
import qualified Data.Map.Strict as Map
import Debug.Trace

import DMSUtil
import ParseParams


sanitizeUnknownAttributes a = 
  map (\(A.UnknownAttribute k v) -> (DTL.unpack k, DTL.unpack v)) a
  
treatNode n =
  let sane = (sanitizeUnknownAttributes . D.nodeAttributes) n
      id = D.nodeID n
  in
   case (lookup "task" sane) of
    Just task -> Left (id, sane)
    Nothing   -> Right (id, sane)

treatEdge e =
  let from = D.fromNode e
      to = D.toNode e
      atts = (sanitizeUnknownAttributes . D.edgeAttributes) e
  in (to, from, atts)

reconcile tasks params =
  map recTask tasks
  where
    recTask task =
      let t = fst task
          lrms = map (classify t) params
          (ins, outs) = partitionEithers $ catMaybes $ lrms
      in (t, ins, outs)
    classify t' p = -- asymptotically quadratic, which is naughty.
      -- trace ("Got " ++ show t' ++ ", " ++ show p) $
        case p of
         (t, dat, [("par", k)])  -> -- input
           if t == t' then Just $ Left (k, dat)
           else Nothing
         (dat, t, [("outpar", k)]) -> -- output
           if t == t' then Just $ Right (k, dat)
           else Nothing
         _ ->
           -- trace ("got " ++ show p) $
           Nothing

accumulate :: Ord a => [(a, b)] -> Map.Map a [b]
accumulate l =
  foldr (\(k, v) m -> Map.insertWith (++) k [v] m) Map.empty l

edgeToTuple e = (D.fromNode e, D.toNode e)

stringifyTask :: (String, [(String, String)], ([(String, String)], [(String, String)])) -> String
stringifyTask (t, params, (ins, outs)) =
  let outString = case outs of
                   [] -> ""
                   [(k, v)] -> v
                   _ ->  (tupleString (keys outs))
  in outString ++ " := " ++ t ++ (tupleString $ (keys $ ins ++ params))
  where
    tupleString l =
      "(" ++ (intercalate ", " l) ++ ")"
    keys :: [(String, String)] -> [String]
    keys l =
      -- map (\(k, v) -> k ++ "=" ++ "\"" ++ v ++ "\"") l
      map (\(k, v) -> k ++ "="  ++ v) l
main = do
  [fn] <- getArgs
  dotText <- L.readFile fn
  let dotGraph = parseDotGraph dotText :: G.DotGraph String
      edges = Data.GraphViz.graphEdges dotGraph
      nodes = Data.GraphViz.graphNodes dotGraph
  hPutStrLn stderr $ "Edges:\n" ++ showNewlineList edges
  hPutStrLn stderr $ "Nodes:\n" ++ showNewlineList nodes
  let tasksOrData = map treatNode nodes
      (tasks, datas) = partitionEithers tasksOrData
      params = map treatEdge edges
      recs = reconcile tasks params
      graphM = Main.accumulate $ map edgeToTuple edges
      (graphG, f1, f2) = graphFromEdges $ map (\(a, l) -> (a, a, l)) $ (Map.toList graphM) 
      tsl = map f1 $ topSort graphG
      taskD = foldr (\(t, ins, outs) m -> Map.insert t (ins, outs) m) Map.empty recs
      -- getTaskTask :: String -> (String, String)
  -- hPutStrLn stderr $ "Original graph" ++ show (dotGraph)
  -- hPutStrLn stderr $ "Parameters: " ++ show(params)
  hPutStrLn stderr $ "\nTasks:" ++ showNewlineList tasks
  hPutStrLn stderr $ "\nRecs: " ++ showNewlineList recs
  hPutStrLn stderr $ "\nGraph: " ++ show graphG
  hPutStrLn stderr $ "\ntopSorted: " ++ show tsl
  let taskList = catMaybes $ map (\(t, _, _) ->
                                   liftM (\x->
                                           let (tn, p) = getTaskTask t tasks
                                           in (tn, p, x))
                                   $ Map.lookup t taskD) tsl
  hPutStrLn stderr $ "taskList: " ++ showNewlineList tsl
  hPutStrLn stdout $ intercalate ";\n" $ map stringifyTask taskList
  where
    getTaskTask t tsks =
      let failVal = []
          rhs = 
            case (lookup t tsks) of
              Nothing -> ("no_task", failVal)
              Just atask -> 
                (case (lookup "task" atask) of
                    Just taskname -> taskname
                    Nothing -> "",
                 case (lookup "pars" atask) of
                   Just pars ->
                     case parseParams pars of
                       Right ps -> ps
                       Left _ ->
                         trace ("Abort failed to parse parameters: "
                                ++ show pars) $
                         failVal
                   Nothing -> failVal)
      in trace ("rhs for " ++ show t ++ " = " ++ show rhs) $
         rhs
