module Kinded (
  ExpF(..),
  Exp,
  eval,
  StatementF(..),
  Statement(..),
  StmPlusF(..),
  StmPlus(..),
  promoteTree,
  enumerateTree,
  applicatiseTreeM,
  collectifyActn,
  collectifyTree
  ) where

{-# LANGUAGE DeriveFunctor, DeriveTraversable #-}

import Control.Monad.State
import Data.Fix
import Data.Foldable
import Data.Traversable
import Data.Map

data ExpF a x = Lit a | Add x x | Sub x x | Mult x x
              deriving (Eq, Ord, Show)
                     
-- We could derive this, with a little pragma magic. I prefer not to
-- do that yet.
instance Functor (ExpF a) where
  fmap _ (Lit a) = Lit a
  fmap f (Add x y) = Add (f x) (f y)
  fmap f (Sub x y) = Sub (f x) (f y)
  fmap f (Mult x y) = Mult (f x) (f y)

type Exp a = Fix (ExpF a)

-- again, we use Data.Fix.cata instead of a home-grown fold.
eval = cata evalF where
  evalF (Lit i) = i
  evalF (Add x y) = x+y
  evalF (Sub x y) = x-y
  evalF (Mult x y) = x*y

type Expression = Exp Integer

data StatementF x = ExpressionStm Expression
                  | CompoundStm x x
                  | AssignStm String Expression
                  | IfStm Expression x x
                  | DoStm Expression x
                  deriving (Eq, Ord, Show, Functor)

type PlusApp a b = StmPlusF a b

type Statement = Fix (StatementF)

type StmNode = StatementF ()

data StmPlusF a x = StmPlusF (a, StatementF x)
                  deriving (Eq, Ord, Show)

type StmPlus a = Fix (StmPlusF a)

newtype StmPlus2 a = StmPlus2 {getStmPlus :: StmPlus a} deriving (Show)

fmapS f (Fix (StmPlusF (a, ExpressionStm e))) =
  Fix $ StmPlusF (f a, ExpressionStm e)
fmapS f (Fix (StmPlusF (a, CompoundStm l r))) =
  Fix $ StmPlusF (f a, CompoundStm (fmapS f l) (fmapS f r))
fmapS f (Fix (StmPlusF (a, AssignStm n v))) =
  Fix $ StmPlusF (f a, AssignStm n v)
fmapS f (Fix (StmPlusF (a, IfStm e l r))) =
  Fix $ StmPlusF (f a, IfStm e (fmapS f l) (fmapS f r))
fmapS f (Fix (StmPlusF (a, DoStm e b))) =
  Fix $ StmPlusF (f a, DoStm e (fmapS f b))
                                
instance Functor StmPlus2 where
  fmap f (StmPlus2 s) = StmPlus2 $ fmapS f s

extractNode (Fix s) = s

promoteTree :: Fix StatementF -> a -> StmPlus a
promoteTree (Fix (ExpressionStm e)) v =
  Fix $ StmPlusF (v, ExpressionStm e)
promoteTree (Fix (CompoundStm l r)) v =
  Fix $ StmPlusF (v, CompoundStm (promoteTree l v) (promoteTree r v))
promoteTree (Fix (AssignStm n val)) v  =
  Fix $ StmPlusF (v, AssignStm n val)
promoteTree (Fix (IfStm e tb eb)) v =
  Fix $ StmPlusF (v, IfStm e (promoteTree tb v) (promoteTree eb v))
promoteTree (Fix (DoStm e s)) v =
  Fix $ StmPlusF (v, DoStm e (promoteTree s v))

extractPlusF (StmPlusF (a, s)) = a

extractPlus = extractPlusF . extractNode

applicatiseTreeM actn t =
  case t of
   Fix (StmPlusF (plus, ExpressionStm e)) -> do
     actn (StmPlusF (plus, ExpressionStm e))
   Fix (StmPlusF (plus, AssignStm name e)) -> do
     actn (StmPlusF (plus, AssignStm name e))
   Fix (StmPlusF (plus, DoStm e t)) -> do
     nt <- applicatiseTreeM actn t
     actn (StmPlusF (plus, DoStm e nt))
   Fix (StmPlusF (plus, CompoundStm l r)) -> do
     nl <- applicatiseTreeM actn l
     nr <- applicatiseTreeM actn r
     actn (StmPlusF (plus, CompoundStm nl nr))
   Fix (StmPlusF (plus, IfStm e tb eb)) -> do
     ntb <- applicatiseTreeM actn tb
     neb <- applicatiseTreeM actn eb
     actn (StmPlusF (plus, IfStm e ntb neb))

subst (StmPlusF (_, e)) a = Fix $ StmPlusF (a, e)

treeEnumActn e = do
  val <- get
  put (succ val)
  return $ subst e val

nullify (ExpressionStm e) = ExpressionStm e
nullify (AssignStm n v)   = AssignStm n v
nullify (DoStm e _)       = DoStm e ()
nullify (CompoundStm _ _) = CompoundStm () ()
nullify (IfStm e _ _)     = IfStm e () ()

processStm :: Ord k => StatementF (Fix (StmPlusF k)) -> k -> Map k [k] -> Map k [k]
processStm s i g =
  case s of
    ExpressionStm _ -> Data.Map.insert i [] g
    AssignStm _ _   -> Data.Map.insert i [] g
    DoStm _ t       -> Data.Map.insert i [extract t] g
    CompoundStm l r -> Data.Map.insert i [extract l, extract r] g
    IfStm _ tb eb   -> Data.Map.insert i [extract tb, extract eb] g 
  where extract = (extractPlusF . extractNode)

collectifyActn (StmPlusF (i, s)) = do
  (map, graph) <- get
  let map' = insert i (nullify s) map
  let graph' = (processStm s i graph)
  put (map', graph') 
  return $ Fix (StmPlusF (i, s))
  
enumerateTree t = evalState (applicatiseTreeM treeEnumActn t) 0

-- We apply the action to the whole tree and then return the state.
collectifyWrap t = do
  applicatiseTreeM collectifyActn t
  get

collectifyTree t = evalState (collectifyWrap t) (empty, empty)


{-

Some examples:


-- e.g.:
let anExp = Fix (Add (Fix (Lit 3)) (Fix (Lit 4))) -- 3 + 4

-- eval (Fix (Add (Fix (Lit 3)) (Fix (Lit 4)))) => 7

stm2 = (Fix (ExpressionStm (Fix (Lit 3))))::Statement

enumerateTree $ promoteTree (Fix $ CompoundStm (Fix (ExpressionStm (Fix (Lit 3)))) (Fix (ExpressionStm (Fix (Lit 3))))) ()

-- promoteTree (Fix (ExpressionStm (Fix (Lit 3)))) () => (ExpStPlus () (Lit 3))

let t = (Fix $ CompoundStm (Fix (ExpressionStm (Fix (Lit 3)))) (Fix (ExpressionStm (Fix (Lit 3)))))

enumerateTree t
-}
